 `define ADD_com   		4'b0000
 `define LOAD_IM_com      4'b0001
 `define SUB_com  		4'b0010
 `define SHIFT_L_com  	4'b0011
 `define SHIFT_R_com 	4'b0100
 `define JMP_com 		4'b0101
 `define SYNC_com		4'b0110	
//
//`define ADDITIONAL_COMMAND   //ǉR}h
`ifdef ADDITIONAL_COMMAND     //ǉR}h
	`define AND_com		4'b0111
	`define OR_com			4'b1000
	`define XOR_com		4'b1001

	
	`define MUL_com		4'b1010
`endif 
 
`define Eq			3'b000
`define Not_Eq		3'b001
`define Always   		3'b010
`define Always_Not     3'b011
`define LT			3'b100
`define GT		      3'b101
`define LTE		      3'b110
`define GTE              3'b111

`define PORT 2'b11
`define R0     2'b01
`define R1	 2'b10
`define ZR     2'b00

module cpu( input clock, reset,
		     input [15:0] port_in_data,
		     input sync,
		     output [15:0] port_out_data);

	localparam integer Loop_counter_width=10;
	localparam integer Stop_count=2**Loop_counter_width -1;

	reg [4:0] pc;
	reg [15:0] port_in_reg;
	reg [Loop_counter_width-1:0] loop_counter;
	reg sync_ff,sync_ff2;

	wire [31:0] rom_data;
	wire [3:0] command=rom_data[31:28];
	wire [1:0] A_address=rom_data [27:26];
	wire [1:0] B_address=rom_data[25:24];
	wire [1:0] write_address=rom_data[23:22];
	wire [2:0] jump_cond=rom_data[21:19];
	wire [2:0] reserved=rom_data[18:16];
	wire [15:0] immediate_data=rom_data[15:0];
	wire loop_flag=reserved[0];


		
	wire [15:0] out_dataA,out_dataB;
	wire [15:0] alu_data;
	

	wire write=1'b1;
	wire [15:0]  Aport_data=command==`LOAD_IM_com 	? immediate_data :port_in_reg;


//loop counter
     always @(posedge clock, posedge reset) begin
		if (reset) loop_counter<=Stop_count;
		else if (loop_flag && loop_counter==Stop_count) loop_counter<=immediate_data;
		else if (loop_flag) loop_counter<=loop_counter -1;
		else loop_counter<=Stop_count;
     end

//sync detection
      always @(posedge clock, posedge reset) begin
		if (reset) sync_ff<=0;
		else 	sync_ff<=sync;
     end
     always @(posedge clock, posedge reset) begin
		if (reset) sync_ff2<=0;
		else 	sync_ff2<=sync_ff;
     end

      wire sinc_pos_edge_detected=sync_ff & ~sync_ff2;

//control of PC
	always @(posedge clock, posedge reset) begin
		if (reset) pc<=0;
		else if (command==`JMP_com && alu_data[0]==1'b1) pc<=immediate_data;//Jump if jump condition is met.
	      else if (command==`SYNC_com) begin//Sync Start Condition
			if (sinc_pos_edge_detected) pc<=pc+1;//Exectute Next Address!
		      else pc<=pc;//Hold until sync pulse is detected.
	      end else if (loop_flag ) begin//Loop Condition
			if (loop_counter==0) pc<=pc+1;//Execute Next Address!
			else pc<=pc;//Hold until loop counter becomes zero.
		end else pc<=pc+1;//else increment PC
	end

//input reg 
      always @(posedge clock,posedge reset) begin
		if (reset) port_in_reg<=0;
		else port_in_reg<=port_in_data;

	end


//ROM
	rom rom(.rom_address(pc),.Data(rom_data));		

//ALU
	 alu  ALU( .command(command),
			    .jump_cond(jump_cond),
			   .portA(out_dataA),
			   .portB(out_dataB),
			   .alu_out(alu_data));


//Register File
	register_file rfile(.clock(clock),
				.write(write),
				.reset(reset),
				.A_address(A_address),
				.B_address(B_address),
				.write_address(write_address),
				.alu_data(alu_data),
				.out_dataA(out_dataA),
				.out_dataB(out_dataB),
				.port_in_data(Aport_data),
				.port_out_data(port_out_data));
				
	

endmodule		


module alu (input [3:0] command,
		    input [2:0] jump_cond,
		    input [15:0] portA,portB,
		   output reg [15:0] alu_out);

	always @* begin
		case (command) 
			`ADD_com :  alu_out=portA+portB;
			`SUB_com :  alu_out=portA-portB;
			`SHIFT_L_com: alu_out=portB >>1;
			`SHIFT_R_com: alu_out=portB <<1;
	`ifdef ADDITIONAL_COMMAND
	//ǉ@R}h
			`MUL_com:      alu_out=portA[7:0]*portB[7:0];//8rbg8rbg=16rbgZ
			`AND_com:      alu_out=portA & portB;
			`OR_com:	  alu_out=portA | portB;
			`XOR_com:     alu_out=portA ^ portB;	
	//ǉI
	`endif
			`JMP_com :   
					case (jump_cond) 
						`Eq : alu_out=portA==portB;
						`Not_Eq:alu_out=portA !=portB;
						`Always :  alu_out=1;
						`Always_Not :alu_out=0;
						`LT :  if (portA < portB) alu_out=1;
							 else			 alu_out=0;
						`GT : if (portA > portB) alu_out=1;
							 else 			alu_out=0;
						default: alu_out=0;
					endcase
			default:  alu_out=portA+portB;


		endcase
     end


endmodule

module rom ( input [4:0] rom_address,
			output [31:0] Data);




parameter [1:0]  Port=`PORT,//IO PORT
			     R0= `R0,
			     R1= `R1,
			     ZR= `ZR;//ZERO REGISTER

parameter [15:0] Zero_Im=16'h0000;

parameter [3:0]  Load_Im=`LOAD_IM_com,
			    Add=`ADD_com,
			    Sub=`SUB_com,
			    Shift_L=`SHIFT_L_com,
			    Shift_R=`SHIFT_R_com,
			    Jmp=`JMP_com,
			    Sync=`SYNC_com
	`ifndef ADDITIONAL_COMMAND //ǉR}h
				;
	`else
				,Mul=`MUL_com,
				And=`AND_com,
				Or =`OR_com,
				Xor=`XOR_com;	
	`endif

parameter [2:0] Eq=`Eq,
			    Not_Eq=`Not_Eq,
			    Always=`Always,
			    Always_Not=`Always_Not,
			   LT=`LT,
			   GT=`GT,
			   LTE=`LTE,
			   GTE=`GTE;	
			    
parameter [2:0] Reserved=3'b000,
			   Loop_Enabled=3'b001;

`define MOV( r0, r1)    {Add,r0,ZR, r1,Eq,Reserved,Zero_Im}
`define CLR( r0)		{ Add,ZR,ZR,r0,Eq,Reserved,Zero_Im}
`define JUMP(num)    {Jmp,ZR,ZR,ZR,Always,Reserved,16'd num}
`define LOAD_IM(r0,num) {Load_Im,Port,ZR,r0,Eq,Reserved,16'd num}
`define ADD(r0,r1,r2) { Add,r0,r1,r2,Eq,Reserved,Zero_Im}
`define SUB(r0,r1,r2) { Sub,r0,r1,r2,Eq,Reserved,Zero_Im}
`define JUMP_IF_LESS_THAN(r0,r1,num) {Jmp,r0,r1,ZR,LT,Reserved,16'd num} 
`define JUMP_IF_GREATER_THAN(r0,r1,num) {Jmp,r0,r1,ZR,GT,Reserved,16'd num} 
`define JUMP_IF_EQ(r0,r1,num) {Jmp,r0,r1,ZR,Eq,Reserved,16'd num} 


`define SHIFT_L(r0,r1) {Shift_L,ZR,r0,r1,Eq,Reserved,Zero_Im}
`define SHIFT_R(r0,r1) {Shift_R,ZR,r0,r1,Eq,Reserved,Zero_Im}
`define SYNC		{Sync,ZR,ZR,ZR,Eq,Reserved,Zero_Im}

//loop command
`define Shift_Right_with_Loop(r0,num)  {Shift_R,ZR,r0,r0,Eq,Loop_Enabled,16'd num }
`define Add_with_loop(r0,r1,num) { Add,r1,r0,r0,Eq,Loop_Enabled,16'd num}  //Use Bport for port_reg

//Micro Field Assignment
// COMMAND 			  [3:0]  4
// A_address			  [1:0]  2 
// B_address			  [1:0]  2
// Write_address			  [1:0]  2
// JUMP_CONDITION         [2:0]  3 
//	EQ
//	NOT_EQ
//   ALWAYS
//   ALWAYS_NOT
//   LT  
//   GT
//   LTE
//   GTE
// others					[2:0]    3
//							[0:0] loop_flag		 
// union 
//	{
//		 JUMP_ADDRESS	    [4:0];
//		 IMMEDIATE			    [15:0];
//          LOOP_COUNTER          [15:0];
//  }
 `ifdef ADDITIONAL_COMMAND
	
	`define MUL(r0,r1,r2) { Mul,r0,r1,r2,Eq,Reserved,Zero_Im}

		function [31:0]romdata;
       	 input [4:0] address;
                case (address)			//Commad  A_addr, B_addr, Write_addr, Jmp_cond,Reserved, IM
                        0:  romdata =`LOAD_IM(R0,100);//RO=100
			     1:  romdata =`MUL(Port,R0,Port);//Port_reg<=R0*|[gin
			     2:  romdata =`JUMP(1);		// 1ԒnɃWv
		  	     default    : romdata=0;
                endcase
 endfunction

`else
 function [31:0]romdata;
        input [4:0] address;
                case (address)			//Commad  A_addr, B_addr, Write_addr, Jmp_cond,Reserved, IM
                        0:  romdata =`LOAD_IM(R0,1);
			     1:  romdata =`CLR(Port);		
                        2:  romdata =`Add_with_loop(Port,R0,100);
		  	     3:  romdata =`CLR(Port);
			     4:  romdata =`LOAD_IM(R0,2);		
                        5 : romdata =`Add_with_loop(Port,R0,50);
			     6:  romdata =`CLR(Port);	
 			     7:  romdata =`LOAD_IM(R0,4);		
                        8 : romdata =`Add_with_loop(Port,R0,25);
			     9 : romdata =`LOAD_IM(Port,75);
			    10: romdata =`Add_with_loop(Port,ZR,25);
  			    11:  romdata =`LOAD_IM(Port,51);
			    12: romdata =`SYNC;	
			    13: romdata =`LOAD_IM(R1,44000);
			     14: romdata =`JUMP_IF_LESS_THAN(R1,Port,1);
                        15 : romdata =`SHIFT_R(Port,Port);
			     16: romdata =`JUMP(14);
                    //    3 : romdata =`LOAD_IM(R0,100);
                   //     4 : romdata =`LOAD_IM(R1, 62444);
                   //     5 : romdata =`ADD(R0,R1,Port);
                   //     6 : romdata =`LOAD_IM(Port,10000);
                   //     7 : romdata =`ADD(R0,R1,R0);
                   //     8 : romdata =`SUB(R0,R1,R0);
                  //      9 : romdata =`ADD(Port,R0,R0);
                  //      10 : romdata =`JUMP_IF_LESS_THAN(R0,R1,0);
                   //     11 : romdata =`JUMP_IF_EQ(R0,R1,3);
                        17 : romdata =0;
                        18 : romdata =0;
                        19 : romdata =0;
                        20 : romdata =0;
			     default    : romdata=0;
                endcase
 endfunction
 `endif
 assign Data = romdata(rom_address);


endmodule


module register_file(input clock,write,reset,
					 input [1:0] A_address,
					 input [1:0] B_address,
					 input [1:0] write_address,
					 input [15:0] alu_data,port_in_data,
					 output [15:0] out_dataA, out_dataB,port_out_data);

   
	reg [15:0] port_reg,r0,r1;
	
	
	always @(posedge clock,posedge reset) begin
		if (reset) begin
				port_reg<=0;
				r0<=0;
				r1<=0;
			end else if (write) begin
				case (write_address)
					`PORT: port_reg<=alu_data;
					`R0: r0<=alu_data;
					`R1: r1<=alu_data;
				//	default r0<=alu_data;
				endcase
		end
	end

	assign out_dataA=A_address==`PORT ? port_in_data :
					 A_address==`R0 ? r0 :
					 A_address==`R1 ?  r1 : 0;

					
   	assign out_dataB=B_address==`PORT ? port_reg :
					 B_address==`R0 ? r0 :
					 B_address==`R1 ? r1 : 0;


	assign port_out_data=port_reg;

endmodule